#include <iostream>
#include <stdexcept>
#include <ruby.h>

#include "library.h"
#include "rubycpp.h"
#include "redirect.h"

namespace EMBEDRUBY {
using std::cout;
using std::endl;
using std::ostream;

RUBY_CPP::Objects *objects = 0;

/*
purpose:
an instance of the ruby class "View"

functions:
*	invoke functions on the instance.
*/
class Impl {
private:
	View *parent;
	VALUE self;
public:
	Impl(View *parent) : parent(parent) {
		/*
		What is going on?
		
		Explanation: todo
		*/
		Redirect::SetView(parent);
		self = RUBY_CPP::New("RubyView");
		objects->Register(self);
	}
	~Impl() {
		objects->Unregister(self);
	}
	void insert() {
		RUBY_CPP::Funcall(self, rb_intern("insert"), 0); 

		/*
		This is intented to provoke a segfault.

		First we unregister our object and running GC.
		Next we try to access the "dead" object, this
		should result in a segfault.

		objects->Unregister(self);
		RUBY_CPP::Funcall(rb_cObject, rb_intern("sweep"), 0); 
		RUBY_CPP::Funcall(self, rb_intern("insert"), 0); 
		*/
	}
};

View::View() {
	cout << "view.ctor: enter" << endl;
	pimpl = new Impl(this);
	cout << "view.ctor: leave" << endl;
}

View::~View() {
	cout << "view.dtor: enter" << endl;
	delete pimpl;
	cout << "view.dtor: leave" << endl;
}

void View::insert() {
	cout << "view.insert: enter" << endl;
	pimpl->insert();
	cout << "view.insert: leave" << endl;
}

void View::repaint() {
	cout << "view.repaint: hello" << endl;
}

/*
purpose:
ensure that ruby is only started/stopped once!

todo: 
*	use a singleton for this
*/
#ifdef EMBEDRUBY_SWIG
	const bool use_swig = EMBEDRUBY_SWIG;
#else
	const bool use_swig = false;
#endif

static bool lib_running = false;

void lib_begin() {
	if(lib_running)
		return;
	
	// initialize ruby itself
	ruby_init();
	ruby_init_loadpath();

	// setup our own environment
	objects = new RUBY_CPP::Objects;
	Redirect::RubyInit();
	RUBY_CPP::Require("test");

	lib_running = true;
	cout << "aeditorlib: started (" << 
		(use_swig ? "with" : "no") << "-swig)" << endl;
}

void lib_end() {
	if(lib_running == false)
		return;
	
	// teardown (reverse initialization order)
	delete objects;
	ruby_finalize();

	lib_running = false;
	cout << "aeditorlib: stopped" << endl;
}

} // end of namespace EMBEDRUBY
